/* icmp.c
	print out the contents of an icmp packet
*/

#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <stdio.h>
#include <time.h>
#include "colorprint.h"
#include "csum.h"
#include "ip.h"

void printICMPpacket (unsigned char *packet, int packlen);
void printts(unsigned long timestamp);

void printICMPpacket (unsigned char *packet, int packlen) {
	unsigned char *nextbyte = packet;
	unsigned short *nextword = (unsigned short *)packet;

	unsigned char type, code;
	unsigned short check;

	HEAD printf("ICMP HEADER:") NORMAL;
	if(packlen < 4) {
		WARN printf("ICMP header too small.  Giving up\n") NORMAL;
		return;
	}
	type = *(nextbyte++);
	code = *(nextbyte++);
	nextword++;

	check = *nextword;
	*nextword = 0;
   TEXT printf("ICMP Checksum:");
   PRINTCSUM(check,csum((unsigned short *)packet,packlen)) NORMAL;

	nextword++;
	nextbyte+=2;
	switch(type) {
	case 8:
	case 0:
		if(type) {
			TEXT printf("Echo request:");
		} else {
			TEXT printf("Echo reply:");
		}
		TEXT printf("code =");
		TESTPRINTu(code==0,code,"should be 0") NORMAL;

		if(packlen < 8) {
			WARN printf("Packet truncated") NORMAL;
			break;
		}
		TEXT printf("Id: %04X",ntohs(*nextword)) NORMAL;
		nextword++;
		TEXT printf("SeqNo: %u",ntohs(*nextword)) NORMAL;
		nextword++;
		nextbyte+=4;
		if(packlen > 8) {
			HEAD printf("Packet data:");
			hexlprint(nextbyte,packlen-8);
		} else {
			TEXT printf("No data in packet") NORMAL;
		}
		break;
	case 3:
		TEXT printf("Destination"); 
		switch(code) {
		case 0:
			TEXT printf("net unreachable") NORMAL;
			break;
		case 1:
			TEXT printf("host unreachable") NORMAL;
			break;
		case 2:
			TEXT printf("protocol unreachable") NORMAL;
			break;
		case 3:
			TEXT printf("port unreachable") NORMAL;
			break;
		case 4:
			TEXT printf("needs fragmentation and DF is set") NORMAL;
			break;
		case 5:
			TEXT printf("unreachable.  Source route failed") NORMAL;
			break;
		case 6:
			TEXT printf("network unknown") NORMAL;
			break;
		case 7:
			TEXT printf("host unknown") NORMAL;
			break;
		case 8:
			TEXT printf("unreachable.  Source host isolated") NORMAL;
			break;
		case 9:
			TEXT printf("network prohibits communication") NORMAL;
			break;
		case 10:
			TEXT printf("host prohibits communication") NORMAL;
			break;
		case 11:
			TEXT printf("network unreachable for type of service") NORMAL;
			break;
		case 12:
			TEXT printf("host unreachable for type of service") NORMAL;
			break;
		default:
			TEXT printf("unreachable. ");
			WARN printf("Unknown code %u",code) NORMAL;
			break;
		}
		nextword+=2;
		nextbyte+=4;
		if(packlen <= 8) {
			WARN printf("Original packet lost") NORMAL;
		} else {
			HEAD printf("Original packet was:") NORMAL;
			doIPpacket(nextbyte,packlen-8);
		}
		break;
	case 4:
		TEXT printf("Source quench: code =") NORMAL;
		TESTPRINTu(code==0,code,"should be 0") NORMAL;
		nextword+=2;
		nextbyte+=4;
		if(packlen <= 8) {
			WARN printf("Original packet lost") NORMAL;
		} else {
			HEAD printf("Original packet was:") NORMAL;
			doIPpacket(nextbyte,packlen-8);
		}
		break;
	case 5:
		TEXT printf("Redirect datagrams");
		switch(code) {
		case 0:
			TEXT printf("for the network");
			break;
		case 1:
			TEXT printf("for the host");
			break;
		case 2:
			TEXT printf("for the type of service and network");
			break;
		case 3:
			TEXT printf("for the type of service and host");
			break;
		default:
			WARN printf("(unknown code %u)",code);
			break;
		}
		if(packlen <= 4) {
			WARN printf("(incomplete packet - desitnation unknown)") NORMAL;
		} else {
			struct hostent *host;
			struct in_addr addr;
 
			addr.s_addr = *(unsigned long *)nextbyte;
			host = gethostbyaddr((char *)&addr,sizeof(struct in_addr),AF_INET);
			TEXT printf("to %u.%u.%u.%u",*nextbyte,*(nextbyte+1),*(nextbyte+2),
							*(nextbyte+3));
			if(host == NULL) {
				WARN printf("not found:");
				switch(h_errno) {
				case HOST_NOT_FOUND:
					WARN printf("Unknown host") NORMAL;
					break;
				case NO_ADDRESS:
					WARN printf("no address associated") NORMAL;
					break;
				case NO_RECOVERY:
					WARN printf("nameserver error") NORMAL;
                        break;
				case TRY_AGAIN:
					WARN printf("temporoary error.  I should try again, but cba")
						NORMAL;
					break;
				default:
					WARN printf("errno: %d\n",h_errno) NORMAL;
					break;
				}
			} else {
				OK printf("(%s)",host->h_name) NORMAL;
			}
			endhostent();
		}
		nextword+=2;
		nextbyte+=4;
		if(packlen <= 8) {
			WARN printf("Original packet lost") NORMAL;
		} else {
         HEAD printf("Original packet was:") NORMAL;
         doIPpacket(nextbyte,packlen-8);
		}
      break;		
	case 11:
		TEXT printf("Time exceeded: ");
		switch(code) {
		case 0:
			TEXT printf("time to live exceeded in transit") NORMAL;
			break;
		case 1:
			TEXT printf("fragment reassembly time exceeded") NORMAL;
			break;
		default:
			WARN printf("unknown code %u",code) NORMAL;
			break;
		}
      nextword+=2;
      nextbyte+=4;
      if(packlen <= 8) {
         WARN printf("Original packet lost") NORMAL;
      } else {
         HEAD printf("Original packet was:") NORMAL;
         doIPpacket(nextbyte,packlen-8);
      }
      break;
	case 12:
		TEXT printf("Parameter problem: ");
		switch(code) {
		case 0:
			TEXT printf("misc problem");
			break;
		case 2:
			TEXT printf("required option missing");
			break;
		default:
			WARN printf("unknown code %u",code);
			break;
		}
		TEXT printf("at offset %u",*nextbyte) NORMAL;
      nextword+=2;
      nextbyte+=4;
      if(packlen <= 8) {
         WARN printf("Original packet lost") NORMAL;
      } else {
         HEAD printf("Original packet was:") NORMAL;
         doIPpacket(nextbyte,packlen-8);
      }
		break;
	case 13:
	case 14:
		if(type == 13) {
			TEXT printf("Timestamp request: ");
		} else {
			TEXT printf("Timestamp reply: ");
		}
      TEXT printf("code =");
      TESTPRINTu(code==0,code,"should be 0") NORMAL;

		if(packlen < 6) {
			WARN printf("Lost rest of packet!") NORMAL;
			break;
		}
		TEXT printf("Id: %04X",ntohs(*nextword));
		nextword++; nextbyte+=2;
		if(packlen < 8) {
			WARN printf("Lost rest of packet!") NORMAL;
			break;
		}
		nextword++; nextbyte+=2;
		
		if(packlen < 20) {
			WARN printf("Lost some or all of timestamps") NORMAL;
			break;
		}
		TEXT printf("Originate time:");
		printts(*(unsigned long *)nextword);
		nextword+=2;nextbyte+=4;
		TEXT printf("Recieve time:");
		printts(*(unsigned long *)nextword);
		nextword+=2;nextbyte+=4;
		TEXT printf("Transit time:");
		printts(*(unsigned long *)nextword);
		nextword+=2;nextbyte+=4;
		if(packlen > 20) {
			WARN printf("Extra crap on end of packet skipped") NORMAL;
		}
		break;
	case 15:
	case 16:
		if(type == 15) {
			WARN printf("Information request: ") NORMAL;
		} else {
			WARN printf("Information reply: ") NORMAL;
		}
      TEXT printf("code =");
      TESTPRINTu(code==0,code,"should be 0") NORMAL;

      if(packlen < 8) {
         WARN printf("Packet truncated") NORMAL;
         break;
      }
      TEXT printf("Id: %04X",ntohs(*nextword)) NORMAL;
		nextword++;
      TEXT printf("SeqNo: %u",ntohs(*nextword++)) NORMAL;
		nextword++;
		if(packlen > 8) {
			WARN printf("Skipping trailing garbage") NORMAL;
		}
		break;
	case 17:
	case 18:
		if(type == 17) {
			TEXT printf("Address mask request: ") NORMAL;
		} else {
			TEXT printf("Address mask reply: ") NORMAL;
		}
		if(packlen < 8) {
			WARN printf("Packet contains nothing useful") NORMAL;
			break;
		}
      TEXT printf("Id: %04X",ntohs(*nextword)) NORMAL;
		nextword++;
      TEXT printf("SeqNo: %u",ntohs(*nextword)) NORMAL;
		nextword++;
		nextbyte+=4;
		TEXT printf("Address mask: %u.%u.%u.%u",*nextbyte,*(nextbyte+1),
						*(nextbyte+2), *(nextbyte+3)) NORMAL;

      if(packlen > 12) {
         WARN printf("Skipping trailing garbage") NORMAL;
      }

		break;
	default:
		WARN printf("Unknown type %u, code %u",type,code) NORMAL;
		break;
	}

	return;
}

void printts(unsigned long timestamp) {
	struct tm *time;
	time_t timet;

	timet = ntohl(timestamp);
	time = localtime(&timet);

	if(time->tm_year) {
		WARN printf("%d years",time->tm_year);
	}
	if(time->tm_mon) {
		WARN printf("%d months",time->tm_mon);
	}
	if(time->tm_mday) {
		WARN printf("%d days",time->tm_mday);
	}
	TEXT printf("%d:%d:%d since midnight",time->tm_hour,time->tm_min,
					time->tm_sec) NORMAL;
}
