/* main.c
	main program code for packprint.  Uses libpcap to grab packets one
	at a time for a file
*/

#include <pcap.h>
#include "ip.h"
#include "arp.h"
#include "icmp.h"
#include "colorprint.h"

void callback(u_char *user, const struct pcap_pkthdr *headers,
				  const u_char *data);
void usage(char *pname);

int main(int argc, char **argv) {
	pcap_t *pcap;
	char ebuf[4096];
	int p;

	if(argc < 2) {
		usage(argv[0]);
		return 1;
	}

	while(++argv, --argc) {
		fprintf(stderr,"Processing file %s\n",*argv);
		pcap = pcap_open_offline(*argv, ebuf);
		if(pcap == NULL) {
			fprintf(stderr,"Couldn't read file: %s\n",ebuf);
			continue;
		}
		
		p = pcap_dispatch(pcap, 0, callback, (u_char *)pcap);
		if(p == -1) {
			pcap_perror(pcap, "pcap_dispatch");
		}
		
		pcap_close(pcap);
	}
	return 0;
}

void usage(char *pname) {
	fprintf(stderr,"Usage: %s file1 file2 .... filen\n",pname);
	fprintf(stderr,"Where file1.... are files produced by eg. tcpdump\n");
}

void callback(u_char *pcap, const struct pcap_pkthdr *headers,
				  const u_char *data) {
	register int i;
	register u_short t;
	int length;
	
	if(headers->caplen < headers->len) {
		WARN printf("Woah!  You only snarfed %u of %u bytes!",headers->caplen,
						headers->len) NORMAL;
		length = headers->caplen;
	} else {
		length = headers->len;
	}

	switch(pcap_datalink((pcap_t *)pcap)) {
	case DLT_NULL:          /* no link-layer encapsulation (loopback) */
		t = *(u_short *)data;
		data+=2;
		if(t) {
			WARN printf("Uhm.  %u != 0 AICMFP",t) NORMAL;
		}
		t = *(u_short *)data;
		data+=2;
		break;
	case DLT_EN10MB:        /* Ethernet (10Mb) */
		HEAD printf("ETHERNET HEADER:") NORMAL;
		TEXT printf("Source MAC address: ");
		for(i=0;i<6;i++) {
			printf("%02X",*(data++));
			if(i<5)printf(":");
			else {NORMAL;}
		}
		TEXT printf("Destination MAC address: ");
		for(i=0;i<6;i++) {
			printf("%02X",*(data++));
			if(i<5)printf(":");
			else {NORMAL;}
		}
		t = *(u_short *)data;
		data+=2;
		break;
	case DLT_RAW:
		t = 0x0008;
		break;
	default:
		WARN printf("Unknown link layer type %u!",
						pcap_datalink((pcap_t *)pcap)) NORMAL;
		t = 0;
		break;
	}
	TEXT printf("Ethertype:");  /* TODO: add some more from RFC 1700 */
	switch(t) {
	case 0x0008:
		OK printf("IP") NORMAL;
		doIPpacket(data, length);
		break;
	case 0x0608:
		OK printf("ARP") NORMAL;
		doARPpacket(data,length);
		break;
	default:
		WARN printf("Unknown %04X",t) NORMAL;
		break;
	}
   
   printf("\n");
	return;
}
