/* tcp.c
 * prints out the contents of a TCP packet header
 */

#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include "colorprint.h"
#include "csum.h"

#define DIE(x) if(packlen<x){WARN printf("Reached end of packet")NORMAL;return;}

void printTCPpacket (unsigned char *packet, int packlen, unsigned char *iphdr);

void printTCPpacket (unsigned char *packet, int packlen, unsigned char *iphdr){
	unsigned char *nextbyte = (unsigned char *)packet;
	unsigned short *nextword = (unsigned short *)packet;
	unsigned long *nextlong = (unsigned long *)packet;
	unsigned short check, check2;
	unsigned int header_len, octets_left;

	struct servent *port;

	HEAD printf("TCP HEADER:") NORMAL;

	if(packlen < 4) {
		WARN printf("Packet contains nothing helpful.  Bailing out") NORMAL;
		return;
	}	

	port = getservbyport(*nextword,"tcp");
	TEXT printf("Source port: %u",ntohs(*nextword));
	if(port) {
		OK printf("(%s)",port->s_name) NORMAL;
	} else {
		WARN printf("(unknown)") NORMAL;
	}
	nextword++; nextbyte+=2;

	port = getservbyport(*nextword,"tcp");
	TEXT printf("Destination port: %u",ntohs(*nextword));
	if(port) {
		OK printf("(%s)",port->s_name) NORMAL;
	} else {
		WARN printf("(unknown)") NORMAL;
	}
	nextlong++; nextword++; nextbyte+=2;
	endservent();

	DIE(8);
	TEXT printf("Sequence number: %u",ntohl(*nextlong)) NORMAL;
	nextlong++; nextword+=2; nextbyte+=4;

	DIE(12);
	TEXT printf("Ack number: %u",ntohl(*nextlong)) NORMAL;
	nextlong++; nextword+=2; nextbyte+=4;

	
	DIE(14);
	header_len = ((*nextbyte)&0xF0) >> 2; /* >> 4, multiply by 4 to get bytes */
	TEXT printf("Header length:");
	TESTPRINTu(header_len > 19,header_len,"should be at least 20") NORMAL;
	TEXT printf("Reserved:");
	check = (ntohs(*nextword)&0x0FC0) >> 6;
	TESTPRINTu(!check,check,"should be 0") NORMAL;
	nextbyte++;
	TEXT printf("Flags:");
	TESTPRINTs(!((*nextbyte)&0x20),"   ","URG");
	TESTPRINTs(!((*nextbyte)&0x10),"   ","ACK");
	TESTPRINTs(!((*nextbyte)&0x08),"   ","PSH");
	TESTPRINTs(!((*nextbyte)&0x04),"   ","RST");
	TESTPRINTs(!((*nextbyte)&0x02),"   ","SYN");
	TESTPRINTs(!((*nextbyte)&0x01),"   ","FIN");
	NORMAL;
	nextword++; nextbyte++;

	DIE(16);
	TEXT printf("Window: %u bytes",ntohs(*nextword)) NORMAL;
	nextlong++; nextword++; nextbyte+=2;

	DIE(20);
	check = *nextword;
	*nextword=0;  /* zero checksum to compute what it should be */
	check2 = tcpcsum(iphdr,(unsigned short *)packet,packlen);
	*nextword = check; /* and restore it afterwards */
	TEXT printf("TCP Checksum:");
	PRINTCSUM(check,check2) NORMAL;
	nextword++; nextbyte+=2;

	TEXT printf("Urgent Pointer: %u",ntohs(*nextword)) NORMAL;
	nextlong++; nextword++; nextbyte+=2;

	if(header_len == 20) {
		TEXT printf("No TCP options") NORMAL;
		return;
	}

	DIE(header_len);
	HEAD printf("TCP OPTIONS:") NORMAL;

	octets_left = header_len-20; check = 0;

	while(octets_left > 0 && !check) {
		switch(*nextbyte) {
		case 0:
			TEXT printf("End of options") NORMAL;
			check = 1;
			nextbyte++; octets_left--;
			break;
		case 1:
			TEXT printf("Nop") NORMAL;
			nextbyte++; octets_left--;
			break;
		case 2:
			TEXT printf("MSS: ( length =");
			nextbyte++;
			TESTPRINTu(*nextbyte == 4,*nextbyte,"should be 4");
			nextbyte++;
			TEXT printf("): %u",(((unsigned short)*nextbyte)<<8)+*(nextbyte+1));
			nextbyte+=2;
			octets_left-=4;
			NORMAL;
			break;
		case 8:
			TEXT printf("Timestamp") NORMAL;
			nextbyte++;
         octets_left -= (*nextbyte);
         nextbyte += (*nextbyte);
         break;
		default:
			WARN printf("Unknown option #%u (length",*(nextbyte++));
			WARN printf("%u)",*nextbyte) NORMAL;
			octets_left -= (*nextbyte);
			nextbyte += (*nextbyte);
			break;
		}
	}
	if(check) {
		TEXT printf("Padding: ");
		while(octets_left > 0) {
			if(*nextbyte) {
				RED printf("%02X",*nextbyte);
			} else {
				GRAY printf("%02X",*nextbyte);
			}
			octets_left--; nextbyte++;
		}
	}
	octets_left=packlen-header_len;
	HEAD printf("TCP packet data:");
	hexlprint(nextbyte, octets_left);
	return;
}
